import tomllib
from pathlib import Path

import cattr
import tomlkit
from dblocks_core import exc
from dblocks_core.config.config import logger
from rich import console, prompt

from dblocks_plugin.jahero_plugin import plug_model

CONFIG_FILE_NAME = "dblocks-jahero-plugin.toml"


def load_config(from_dir: Path) -> plug_model.PluginConfig:
    # check that the file exists
    config_file = from_dir / CONFIG_FILE_NAME
    if not config_file.is_file():
        write_default_config(config_file)
        raise exc.DConfigError(f"config not found: {config_file}")

    # load it
    logger.info(f"read config from {config_file}")
    data = tomllib.loads(config_file.read_text(encoding="utf-8"))
    cfg = cattr.structure(data, plug_model.PluginConfig)
    return cfg


def write_default_config(file: Path):
    cnsl = console.Console()
    cnsl.print("Config file not found", style="bold red")
    cnsl.print(file.as_posix(), style="green")
    if (
        prompt.Prompt().ask(
            "Do you want to create file with default config? [Y/n]", default="Y"
        )
        != "Y"
    ):
        return

    cfg = make_default_config()
    data = cattr.unstructure(cfg)
    string = tomlkit.dumps(data, sort_keys=True)
    file.write_text(string, encoding="utf-8")


def make_default_config() -> plug_model.PluginConfig:
    return plug_model.PluginConfig(
        replacements=[
            plug_model.Replacement(
                replace_from="^EP_CVM(.*)$",
                replace_to=(r"ED1_CVM\1"),
            ),
            plug_model.Replacement(replace_from="^EP_(.*)$", replace_to=(r"ED0_\1")),
            plug_model.Replacement(replace_from="^AP_(.*)$", replace_to=(r"AD0_\1")),
            plug_model.Replacement(replace_from="^VP_(.*)$", replace_to=(r"VD0_\1")),
        ],
        cc=plug_model.ConditionalCreate(
            max_files=50,
            conditionals=[
                plug_model.ConditionalCreate(
                    path="DB/Teradata/01-copy-source-ddl-tbl",
                    scenario=plug_model.DROP,
                )
            ],
        ),
    )



from pathlib import Path

from attrs import field, frozen
from dblocks_core.config.config import logger

CREATE = "create"
DROP = "drop"
CLEANUP = "cleanup"
DROP_ONLY = "drop-only"
VALID_SCENARIOS = [CREATE, DROP, DROP_ONLY, CLEANUP]


def _assert_valid_cc_scenario(self, attribute, value):
    if not isinstance(value, str):
        err = ValueError(f"string expected, got: {str(type(value))}")
        logger.error(err)
        raise err

    if value not in VALID_SCENARIOS:
        err = ValueError(
            f"invalid value: {attribute}: {value=}, expected one of: {VALID_SCENARIOS=}"
        )
        logger.error(err)
        raise err


def _assert_not_empty_string(self, attribute, value):
    if not isinstance(value, str):
        err = ValueError(f"string expected, got: {str(type(value))}")
        logger.error(err)
        raise err

    if value == "":
        err = ValueError(f"not empty string was expected, got: {value=}")
        logger.error(err)
        raise err
    if " " in value:
        err = ValueError(f"string with no white space expected, got: {value=}")
        logger.error(err)
        raise (err)


@frozen
class Replacement:
    replace_from: str = field(validator=_assert_not_empty_string)
    replace_to: str = field(validator=_assert_not_empty_string)


@frozen
class ConditionalCreatePath:
    path: Path  # dblocks has registered global converter for this data type, we do not have to use str
    scenario: str = field(default="drop")
    input_encoding: str = field(default="utf-8")
    output_encoding: str = field(default="utf-8")


@frozen
class ConditionalCreate:
    max_files: int = field(default=20)
    conditionals: list[ConditionalCreatePath] = field(factory=list)


@frozen
class PluginConfig:
    replacements: list[Replacement] = field(factory=list)
    cc: ConditionalCreate



import re
import stat
from pathlib import Path
from textwrap import dedent

from dblocks_core import exc
from dblocks_core.config import config
from dblocks_core.config.config import logger
from dblocks_core.deployer import fsequencer
from dblocks_core.model import config_model, plugin_model

from dblocks_plugin.jahero_plugin import plug_config, plug_model


class Dpl(plugin_model.PluginWalker):
    def __init__(self):
        self.batch: fsequencer.DeploymentBatch | None = None

    def before(
        self,
        path: Path,
        environment: str | None,
        cfg: config_model.Config,
        **kwargs,
    ):
        """This function is executed before the walk starts."""
        # get plugin config
        plug_cfg = plug_config.load_config(path)

        # deploy
        pkg_cfg = cfg.packager
        if pkg_cfg.case_insensitive_dirs:
            logger.info("case insensitive search")
            subdirs = case_insensitive_search(path, pkg_cfg.steps_subdir)

            if subdirs is None:
                raise exc.DOperationsError(
                    f"subdir not found: {pkg_cfg.steps_subdir}: in {path}"
                )
            root_dir = path / subdirs
        else:
            logger.warning("case SENSITIVE search")
            root_dir = path / pkg_cfg.steps_subdir

        # sanity check
        if not root_dir.is_dir():
            raise exc.DOperationsError(f"directory not found: {root_dir}")

        # get a log directory in the package
        log_dir = path / "log"
        log_dir.mkdir(exist_ok=True)

        self.batch = fsequencer.create_batch(root_dir=root_dir, tgr=None)
        if len(self.batch.steps) == 0:
            raise exc.DOperationsError("Empty batch.")

        exec_scripts = []
        deployment_dir = root_dir.parent / "_deployment"
        log_dir = path / "log"

        deployment_dir.mkdir(exist_ok=True, parents=True)
        log_dir.mkdir(exist_ok=True, parents=True)

        for step in self.batch.steps:
            logger.info(f"creating step: {step.name}")
            statements, prev_db = [], None

            # skip empty steps
            if len(step.files) == 0:
                logger.error(f"empty step: {step.name}")
                continue

            bteq_file = deployment_dir / (step.name + ".bteq")
            exec_scripts.append(_get_bteq_call(bteq_file, log_dir))

            for f in step.files:
                db = f.default_db

                if db is not None and db != prev_db:
                    statements.append(f"\ndatabase {_get_database(db, plug_cfg)};")
                statements.append(f".run file = '{f.file.absolute().as_posix()}'")

                prev_db = db

            if statements:
                script = _get_header() + "\n".join(statements) + _get_footer()
                bteq_file.write_text(script, encoding="utf-8")
            else:
                logger.error(f"EMPTY STEP: {step.name}")

        runme_file = deployment_dir / "runme.sh"
        script = "#!/bin/bash\n\n" + "\n\n".join(exec_scripts)
        runme_file.write_text(script, encoding="utf-8")
        runme_file.chmod(runme_file.stat().st_mode | stat.S_IEXEC)
        # os.chmod(runme_file, 755)

    def walker(
        self,
        path: Path,
        environment: str | None,
        cfg: config_model.Config,
        **kwargs,
    ):
        pass

    def after(
        self,
        path: Path,
        environment: str | None,
        cfg: config_model.Config,
        **kwargs,
    ):
        pass


def case_insensitive_search(root: Path, subdir: Path) -> Path | None:
    wanted = _path_to_directories(subdir)
    wanted = [s.lower() for s in wanted]
    logger.info(f"searching in: {root}")
    logger.info(f"searching for: {wanted}")
    found_dirs = []

    for i in range(len(wanted)):
        children_dir_names = [
            (d.name.lower(), d.name) for d in root.glob("*") if d.is_dir
        ]
        found = False
        for name_lower, name in children_dir_names:
            if name_lower == wanted[i]:
                found = True
                found_dirs.append(name)
                root = root / name
                break
        if not found:
            return None

    return Path(*found_dirs)


def _path_to_directories(path: Path) -> list[str]:
    elements = []
    curr: Path = path
    prev: Path | None = None

    while curr != prev:
        if curr.name:
            elements.insert(0, curr.name)
        prev = curr
        curr = curr.parent

    return elements


def _get_header() -> str:
    return dedent(
        """
        -----------------------------------------------------------
        .SET SESSION CHARSET 'UTF8'
        .SET WIDTH 65531
        .SET ERRORLEVEL UNKNOWN SEVERITY 8;
        .SET ERROROUT STDOUT;
        .SET MAXERROR 1
        -----------------------------------------------------------
        .RUN FILE='/home/jan/Vaults/o2/logon_prod.sql'
        -----------------------------------------------------------
        SET SESSION DATEFORM=ANSIDATE;
        .SET ERRORLEVEL 3624 SEVERITY 0;            -- collect stats - pro neex. stat - projde
        --.SET ERRORLEVEL 3803 SEVERITY 0;          -- projde create tabulky ktera existuje
        --.SET ERRORLEVEL 3807 SEVERITY 0;          -- projde drop tabulky ktera neexistuje
    """
    )


def _get_footer() -> str:
    return ""


def _get_bteq_call(f: Path, log_dir: Path) -> str:
    name, stem = f.name, f.stem
    log_file = log_dir / f"{stem}.log"
    return dedent(
        f"""
        echo "running {stem}"
        bteq < {f.absolute().as_posix()} &>>{log_file.absolute().as_posix()}
        retval=$?
        if [ $retval -ne 0 ]; then
            echo "===============ERROR================="
            echo "====================================="
            exit $retval
        fi           
    """
    )


def _get_database(db: str, cfg: plug_model.PluginConfig) -> str:
    for replacement in cfg.replacements:
        db = re.sub(
            replacement.replace_from,
            replacement.replace_to,
            db,
            flags=re.I | re.X,
        )
    return db



import re
from pathlib import Path

from dblocks_core import exc
from dblocks_core.config.config import logger
from dblocks_core.model import config_model, plugin_model
from rich.console import Console
from rich.prompt import Prompt

from dblocks_plugin.jahero_plugin import plug_config, plug_model

_CREATE = "create"
_DROP = "drop"
_CLEANUP = "cleanup"
_DROP_ONLY = "drop-only"

_VALID_SCENARIOS = [_CREATE, _DROP, _DROP_ONLY, _CLEANUP]

# regular expressions building blocks
_WS = r"\s+"
_MAYBE_WS = r"\s*"
_SET = "set +|multiset +"
_TABLE = "table"
_IDENTIFIER = "[a-z0-9_.]+"
_AS = "as"
_PATTERN = (
    _CREATE
    + _WS
    + "("
    + _SET
    + ")?"
    + _MAYBE_WS
    + _TABLE
    + _WS
    + "(?P<IDENT>"
    + _IDENTIFIER
    + ")"
)
_PATTERN2 = (
    _CREATE
    + _WS
    + _TABLE
    + _WS
    + "(?P<IDENT>"
    + _IDENTIFIER
    + ")"
    + _AS
    + _WS
    + _IDENTIFIER
)

# regular expressions used to find what is needed
_re_create = re.compile(_PATTERN, re.I)
_re_create2 = re.compile(_PATTERN2, re.I)


# the template inected into SQL files
_TEMPLATE_CREATE = """-----------------------------------------------------------------------
/*--conditional_create--*/
select count(*) from dbc.tablesV 
where tablename='<table>' and databasename=( select database )
having count(*) > 0;

.IF ERRORCODE <> 0 THEN .QUIT 10;
.IF ACTIVITYCOUNT = 0 THEN .GOTO TABLE_DOES_NOT_EXIST;
.REMARK 'tabulka EXISTUJE'
.GOTO KONEC

.LABEL TABLE_DOES_NOT_EXIST
-----------------------------------------------------------------------
"""

_TEMPLATE_DROP = """-----------------------------------------------------------------------
/*--conditional_create--*/
select count(*) from dbc.tablesV 
where tablename='<table>' and databasename=( select database )
having count(*) > 0;

.IF ERRORCODE <> 0 THEN .QUIT 10;
.IF ACTIVITYCOUNT = 0 THEN .GOTO TABLE_DOES_NOT_EXIST;
.REMARK 'tabulka EXISTUJE'
drop table <table>;

.LABEL TABLE_DOES_NOT_EXIST
-----------------------------------------------------------------------
"""


def _load_script(file: Path, input_encoding: str):
    """loads the script, and filters out the header and footer of the conditional create.

    Args:
        file (Path): path to the file
        input_encoding (str): encoding used to load the file

    Returns:
        _type_: _description_
    """
    outLines = []
    with open(file, "r", encoding=input_encoding) as f:
        script = f.read()

        # pokud neobsahuje ".LABEL TABLE_DOES_NOT_EXIST", není tam co řešit
        if ".LABEL TABLE_DOES_NOT_EXIST" not in script:
            return script

        # jinak ale musíme filtrovat
        lines = script.splitlines()
        skipThis = True
        readNext = False

        for line in lines:
            if line == ".LABEL TABLE_DOES_NOT_EXIST":
                readNext = True
                continue
            if readNext:
                skipThis = False
                readNext = False
                continue

            if line == ".LABEL KONEC":
                break

            if skipThis:
                continue

            outLines.append(line)

    return "\n".join(outLines)


def _is_create_script(insides: str, file: str):
    if m := _re_create.search(insides):
        parts = m.group("IDENT").split(".")
        if len(parts) == 1:
            return True, parts[0]
        return True, parts[1]
    elif m := _re_create2.search(insides):
        parts = m.group("IDENT").split(".")
        if len(parts) == 1:
            return True, parts[0]
        return True, parts[1]
    else:
        raise exc.DOperationsError(f"did not find name of the table: {file}")


def _change_to_conditional(
    file: Path,
    output_encoding: str,
    input_encoding: str,
    create_or_drop=_DROP,
):
    try:
        old_script = _load_script(file, input_encoding)
    except UnicodeDecodeError as e:
        raise exc.DOperationsError(f"encoding error: {file.as_posix()}")

    is_create, creates_table = _is_create_script(old_script, file)

    if not is_create:
        return False

    new_script = ""

    if create_or_drop == "cleanup":
        new_script = old_script

    else:
        TEMPLATE = _TEMPLATE_CREATE if create_or_drop == "create" else _TEMPLATE_DROP
        inject = TEMPLATE.replace("<table>", creates_table)
        if create_or_drop == "drop-only":
            old_script = "--only drop the table, do not create\n\n"
        new_script = inject + old_script + "\n" + ".LABEL KONEC"

    # save
    if new_script != old_script:
        logger.info(f"writing: {file.as_posix()}")
        try:
            with open(file, "w", encoding=output_encoding) as f:
                f.write(new_script)
        except Exception as e:
            logger.error(f"check this file, it may have been lost: {file.as_posix()}")


class CC(plugin_model.PluginWalker):
    def __init__(self):
        self.files: list[Path] = []
        self.cc_abs_path: Path | None = None

    def before(
        self,
        path: Path,
        environment: str | None,
        cfg: config_model.Config,
        **kwargs,
    ):
        """This function is executed before the walk starts."""
        self.cc_abs_path = path.absolute()

    def walker(
        self,
        path: Path,
        environment: str | None,
        cfg: config_model.Config,
        **kwargs,
    ):
        """This function is executed for each file we walk through."""
        ext = path.suffix.lower()
        if ext in (".bteq", ".sql"):
            self.files.append(path.absolute())

    def after(
        self,
        path: Path,
        environment: str | None,
        cfg: config_model.Config,
        **kwargs,
    ):
        plug_cfg = plug_config.load_config(path)
        self._conditional_create(plug_cfg)

    def _conditional_create(
        self,
        cfg: plug_model.PluginConfig,
    ):
        """
        Scans a folder for files with .sql or .bteq extension, and tries to insert conditional create header and footer in them.
        Raises a ContentError when the number of files is > max_files.

        Args:
            folder (str): input folder
            scenario (str, optional): ["drop", "create", "cleanup", "drop-only"], defaults to "create".
            input_encoding (str, optional): defaults to "utf-8".
            output_encoding (str, optional): defaults to "utf-8".
            max_files (int, optional): defaults to 20.

        Raises:
            ContentError: raised when number of files is too great, or when one of the files can not be patched.
        """
        files = self.files
        max_files = cfg.cc.max_files
        if len(files) > max_files:
            console = Console()
            console.print(f"Expected to get at max {max_files}, got {len(files)}")
            answer = Prompt.ask("continue? [Y/n]")
            if answer != "Y":
                console.print("Canceled.", style="bold red")
                return
        for cc in cfg.cc.conditionals:
            cc_abs_path = self.cc_abs_path / cc.path
            for f in files:
                if f.is_relative_to(cc_abs_path):
                    _change_to_conditional(
                        f,
                        cc.output_encoding,
                        cc.input_encoding,
                        cc.scenario,
                    )
